#!/usr/bin/python
#
# use spring calibration and assorted constants to scale data to pN vs nm
#
# usage: scale_unfold data_directory [more directories]

VERSION = "0.2"

import scipy.io
import piezo.z_piezo_utils as z_piezo_utils
import calibcant.bump_analyze
import calibcant.config
import numpy
import os, re, time, calendar
import shutil
import data_logger

LOG_SCALED_AVGD_DATA = True
LOG_DIR = '$DEFAULT$/scaled_unfold'
CANTILEVER_CALIBRATION_DIR = calibcant.config.LOG_DIR
FORCED_CANTILEVER_CALIBRATION_FILE = None
#SET_ZERO_POS = "approach"
#SET_ZERO_POS = "retract"
SET_ZERO_POS = "both"

AVERAGE = False
INVERT_Z = False
INVERT_F = False
REMOVE_CANT = False
CALIB_SENSITIVITY = False
ZERO_DEF_FROM_TAILING_N = None

# Prof. Yang's software does not save approach data (so far as I know to date)
# This option allows you to move on without it.
ALLOW_MISSING_APPROACH = False

VERBOSE = 0

def load_pull(basepath) :
    approach_file = "%s_%s" % (basepath,'approach')
    param_file = "%s_%s" % (basepath,'param')

    unlabeled_unfold_data=scipy.io.read_array("%s" % basepath)
    
    if os.path.exists(approach_file):
        use_approach = True
        unlabeled_approach_data=scipy.io.read_array(approach_file)
        approach_data = {}
    else:
        assert ALLOW_MISSING_APPROACH == True, \
            "Missing approach file %s, but ALLOW_MISSING_APPROACH not set." \
            % (approach_file)
        use_approach = False
        unlabeled_approach_data = None
        approach_data = None
    
    unfold_data = {}
    params = {}
    for line in file(param_file) :
        ldata = line.split(":")
        if ldata[0] == "Data fields" :
            data_fields = ldata[1].split()
        elif ldata[0] == "Time" :
            params[ldata[0]] = ldata[1].split()[0]
        elif len(ldata) == 2 :
            argwords = ldata[1].split()
            if len(argwords) == 1 : # single word probably a number
                params[ldata[0]] = float(ldata[1])
            else :
                params[ldata[0]] = ldata[1]
    
    # move data into a dictionary of labeled timeseries
    for i in range(  min(len(data_fields), 
                         unlabeled_unfold_data.shape[1])  ) :
        unfold_data[data_fields[i]] = unlabeled_unfold_data[:,i]
        if unlabeled_approach_data == None:
            pass
        elif len(unlabeled_approach_data.shape) == 1 : # only one point
            approach_data[data_fields[0]] \
                = numpy.array([unlabeled_approach_data[0]])
            approach_data[data_fields[1]] \
                = numpy.array([unlabeled_approach_data[1]])
        else :
            for i in range(  min(len(data_fields),
                                 unlabeled_approach_data.shape[1])  ) :
                approach_data[data_fields[i]] = unlabeled_approach_data[:,i]

    return (approach_data, unfold_data, params)

def list_re_search(list, regexp) :
    "Return list entries matching re"
    reg = re.compile(regexp)
    ret = []
    for item in list :
        if reg.match(item) :
            ret.append(item)
    return ret

def list_calibration_files(timestamp=None, basedir=None):
    if timestamp == None:
        timestamp = time.time()
    if basedir == None:
        basedir = CANTILEVER_CALIBRATION_DIR
    YYYYMMDD = time.strftime("%Y%m%d", time.localtime(timestamp))
    dir = os.path.join(data_logger.normalize_logdir(basedir), YYYYMMDD)
    if not os.path.exists(dir):
        return []
    all_calibfiles = os.listdir(dir) # everything in the directory
    #regexp = "%s.*_analysis_text " % (YYYYMMDD)
    regexp = ".*_analysis_text"
    calibfiles = list_re_search(all_calibfiles, regexp)
    paths = [os.path.join(dir, cf) for cf in calibfiles]
    return paths

def time_from_localtime_string(timestring, format="%Y%m%d%H%M%S"):
    """
    >>> print time.tzname
    ('EST', 'EDT')
    >>> scale_unfold.time_from_localtime_string("19700101", format="%Y%m%d")/3600.0
    5.0
    >>> scale_unfold.time_from_localtime_string("20081231", format="%Y%m%d")
    1230699600
    >>> scale_unfold.time_from_localtime_string("20090101", format="%Y%m%d")
    1230786000
    """
    timestruct = time.strptime(timestring, format)
    timestamp = calendar.timegm(timestruct)
    # Date strings are in localtime, so what we really want is
    # .timelocale, but that doesn't exist.  Workaround to convert
    # timestamp to UTC.
    timestamp += time.timezone # assume no DST
    if bool(time.localtime(timestamp).tm_isdst) == True:
        timestamp -= time.timezone - time.altzone # correct if DST
    assert time.strftime(format, time.localtime(timestamp)) == timestring, "error in time_from_localtime_string, %s != %s" % (time.strftime(format, time.localtime(timestamp)), timestring)
    return timestamp

def calibfile_timestamp(path):
    filename = os.path.basename(path)
    YYYYMMDDHHMMSS = filename[0:14]
    return time_from_localtime_string(YYYYMMDDHHMMSS, format="%Y%m%d%H%M%S")

def find_previous_calibration_file(timestamp=None, basedir=None,
                                   previous_days=1) :
    """
    If timestamp == None, uses current time.
    Warning : brittle! depends on the default data_logger.py save filesystem.
    Renaming files or moving directories will break me.
    """
    #print "calibration timestamp ", timestamp
    if timestamp == None :
        timestamp = time.time()
    if type(timestamp) == type("a string") :
        timestamp = time_from_localtime_string(timestamp, format="%Y%m%d%H%M%S")
    calibfiles = []
    for i in range(previous_days+1):
        calibfiles.extend(list_calibration_files(timestamp-24*3600*i,basedir=basedir))
    assert len(calibfiles) > 0 , \
        "No calibration files in that day range in directory '%s'" % basedir
    calibfiles.sort()
    if VERBOSE >= 5:
        print "calibfiles:\n ",
        print "\n  ".join(calibfiles)
    for i in range(len(calibfiles)) :
        if calibfile_timestamp(calibfiles[i]) > timestamp :
            i -= 1
            break
    if VERBOSE >= 4:
        print calibfiles[i], "is first calib before", time.strftime("%Y/%m/%d-%H:%M:%S", time.localtime(timestamp))
    return os.path.join(dir, calibfiles[i])

def get_k_from_file(calibfile) :
    for line in file(calibfile, "r") :
        var = line.split(" : ")[0]
        if var == "k (N/m)" :
            return float(line.split(" : ")[1].split(" +/- ")[0])
        elif var == "Cantilever k (N/m)           " :
            return float(line.split(" : ")[1].split(" +/- ")[0])
    raise Exception, "No 'k (N/m)' found in file '%s'" % calibfile

def get_a_from_file(calibfile) :
    for line in file(calibfile, "r") :
        var = line.split(" : ")[0]
        if var == "a**2 (V/nm)**2" or var == "photoSensitivity**2 (V/nm)**2" :
            a2 = float(line.split(" : ")[1].split(" +/- ")[0])
            return a2**0.5
    raise Exception, "No 'a**2 (V/nm)**2' found in file '%s'" % calibfile

def scale_zp_array(array, sensitivity, gain, surfPos, invert=False) :
    """
    Convert the output I16 voltages to nm positions
    surfPos is in output units
    """
    # zero surfPos and convert to volts (23768 bits = 2*15 bits = 10 Volts)
    array_volts = (array.astype(numpy.float) - surfPos)/3276.8
    array_nm = array_volts * gain * sensitivity;
    if invert == True:
        array_nm *= -1.0
    if VERBOSE >= 6:
        print "Scaling z piezo bits to Volts with"
        print "  x_V(b) = (b-surf)/3276.8"
        print "    where surf = %g bits" % surfPos
        print "  x_nm(x_V) = x_V * gain * piezo_sensitivity"
        print "    where gain = %g" % gain
        print "    where piezo_sensitivity = %g nm/V" % sensitivity
        if invert == True:
            print "  Inverting"
    if VERBOSE >= 7:
        print "  the range of x_nm is [%g, %g] pN" \
            % (min(array_nm), max(array_nm))
    return array_nm;

def scale_def_array(def_array, zp_array, surfPos, k, photo_sensitivity,
                    invert=False) :
    """
    Convert the input I16 voltages to pN forces
    both arrays are in raw comedi units
    k is the cantilever spring constant in in N/m
    photo_sensitivity is in Vphotodiode/nm_tip (a from calibration)
    """
    # zero and convert to volts (23768 bits = 2*15 bits = 10 Volts)
    avgZeroDef = 0;
    num = 0;
    if ZERO_DEF_FROM_TAILING_N == None:
        # Average for all points off of the surface
        for i in range(len(zp_array)) :
            if zp_array[i] < surfPos :
                avgZeroDef += def_array[i]
                num += 1
    else:
        # avoid protein deflection biasing the average by only taking
        # the last N points.  Drawbacks are arbitratry N and reduced
        # robustness to noise.
        num = int(ZERO_DEF_FROM_TAILING_N)
        avgZeroDef = sum(def_array[-ZERO_DEF_FROM_TAILING_N:])
    if num > 0:
        avgZeroDef /= float(num)
    else:
        avgZeroDef = def_array[-1]
    array_volts = (def_array.astype(numpy.float) - avgZeroDef)/3276.8
    array_nm = array_volts / photo_sensitivity
    array_pN = array_nm * (k * 1000)
    if invert == True:
        array_pN *= -1.0
    if VERBOSE >= 6:
        print "Scaling photodiode deflection bits to pN with"
        print "  x_V(b) = (b-avg_def)/3276.8"
        print "    where avg_def = %g bits (average deflection after surface)"\
            % avgZeroDef
        print "  x_nm(x_V) = x_V / photodiode_sensitivity"
        print "    where photodiode_sensitivity = %g V/nm" % photo_sensitivity
        print "  x_pN(x_nm) = x_nm * (k * 1000)"
        print "    where k = %g N/m" % k
        if invert == True:
            print "  Inverting"
    if VERBOSE >= 7:
        print "  the range of x_pN is [%g, %g] pN" \
            % (min(array_pN), max(array_pN))
    return array_pN;

def remove_cantilever_displacement(scaled_zp_array, scaled_def_array, k):
    """
    Adjust zp array (in nm) to remove the cantilever bending distance.
    This converts a surface-to-unbent-cantilever-tip distance to a
    surface-to-current-cantilever-tip distance (usually more
    interesting).  Returns the new zp array (but it may alter the
    input zp array as well).
    """
    if VERBOSE >= 6:
        print "Removing cantilever displacement (k=%g)" % k
    for i, z, F in zip(range(len(scaled_zp_array)), scaled_zp_array, scaled_def_array):
        scaled_zp_array[i] = scaled_zp_array[i] - F/(k*1e3) # 1e3 to convert N/m to pN/nm
    return scaled_zp_array

def scale_data(approach_data, unfold_data, params, k, a_calib) :
    # Center the z piezo values so 0 corresponds to the surface position
    if approach_data != None:
        if len(approach_data["Deflection_in"]) > 10 :
            ddict = {"approach":{"Deflection input":\
                                     approach_data["Deflection_in"],
                                 "Z piezo output":\
                                     approach_data["Z_piezo_out"]}}
            (non_contact_intercept,non_contact_slope,
             surface_position_approach,contact_slope_approach) \
             = z_piezo_utils.analyzeSurfPosData(ddict, retAllParams=True)
             # slope is in Ybits/Zbits
            slope_approach_V2nm = calibcant.bump_analyze.slope_bitspbit2Vpnm( \
                contact_slope_approach) # TODO, kwargs
        else :
            # not many approach points, so do nothing
            surface_position_approach = 0
            slope_approach_V2nm = 0
    else:
        surface_position_approach = 0
        slope_approach_V2nm = 0

    ddict = {"approach":{"Deflection input":unfold_data["Deflection_in"],
                         "Z piezo output":unfold_data["Z_piezo_out"]}}
    (non_contact_intercept,non_contact_slope,
     surface_position_unfold,contact_slope_unfold) \
        = z_piezo_utils.analyzeSurfPosData(ddict, retAllParams=True)
    slope_unfold_V2nm = calibcant.bump_analyze.slope_bitspbit2Vpnm(contact_slope_unfold) # TODO, kwargs

    if SET_ZERO_POS == "approach" :
        # use approach position for scaling unfold
        assert approach_data != None and len(approach_data) > 10
        surface_position_unfold = surface_position_approach 
        slope_unfold_V2nm = slope_approach_V2nm
    elif SET_ZERO_POS == "unfold" :
        # use unfold position for approach unfold
        surface_position_approach = surface_position_unfold
        slope_approach_V2nm = slope_unfold_V2nm
    
    if CALIB_SENSITIVITY == True:
        slope_approach_V2nm = a_calib
        slope_unfold_V2nm = a_calib
    
    # Convert deflections to pNewton forces, needs raw piezo output arrays
    if approach_data != None:
        approach_data["Deflection_in"] = scale_def_array(
            approach_data["Deflection_in"],
            approach_data["Z_piezo_out"],
            surface_position_approach,k,slope_approach_V2nm, INVERT_F)
    unfold_data["Deflection_in"] = scale_def_array(
        unfold_data["Deflection_in"],
        unfold_data["Z_piezo_out"],
        surface_position_unfold,k,slope_unfold_V2nm, INVERT_F)

    # Convert the Z piezo output voltages to nm positions
    if approach_data != None:
        approach_data["Z_piezo_out"] = scale_zp_array(
            approach_data["Z_piezo_out"],
            params["Z piezo sensitivity (nm/Vp)"],
            params["Z piezo gain (Vp/Vo)"],
            surface_position_approach, INVERT_Z)
    unfold_data["Z_piezo_out"] = scale_zp_array(unfold_data["Z_piezo_out"],
                                   params["Z piezo sensitivity (nm/Vp)"],
                                   params["Z piezo gain (Vp/Vo)"],
                                   surface_position_unfold, INVERT_Z)

    if REMOVE_CANT == True:
        if approach_data != None:
            approach_data["Z_piezo_out"] = \
                remove_cantilever_displacement(approach_data["Z_piezo_out"],
                                               approach_data["Deflection_in"],
                                               k)
        unfold_data["Z_piezo_out"] = \
            remove_cantilever_displacement(unfold_data["Z_piezo_out"],
                                           unfold_data["Deflection_in"],
                                           k)
    
    return (approach_data, unfold_data)

def average_unfold_data(unfold_data, params) :
    n = params['Points per step']
    un = len(unfold_data['Deflection_in'])
    avg_zp = numpy.zeros((un/n,), dtype=numpy.float)
    avg_def = numpy.zeros((un/n,), dtype=numpy.float)
    if VERBOSE >= 3:
        print "Average %d reads per step for %d steps" % (n, un/n)
    for i in range(len(avg_zp)) :
        avg_zp[i] = unfold_data['Z_piezo_out'][i*n:(i+1)*n].mean()
        avg_def[i] = unfold_data['Deflection_in'][i*n]
    return {'Z_piezo_out':avg_zp, 'Deflection_in':avg_def}

def read_scale_avg(pull_basepath, plotVerbose=False, average=True) :
    approach_data, unfold_data, params = load_pull(pull_basepath)
    if FORCED_CANTILEVER_CALIBRATION_FILE != None:
        calibfile = FORCED_CANTILEVER_CALIBRATION_FILE
    else:
        calibfile = find_previous_calibration_file(params["Time"])
    if VERBOSE >= 2:
        print "Calibfile", calibfile
    params['Cantilever calibration file'] = calibfile
    k = get_k_from_file(calibfile) # cantilever spring constant
    if CALIB_SENSITIVITY == True:
        a_calib = get_a_from_file(calibfile) # calibrated photodiode sensitivity
    else:
        a_calib = None
    if VERBOSE >= 3:
        print "Spring const", k, "N/m"
        print "Sensitivity", a_calib, "V/nm"
    approach, unfold = scale_data(approach_data, unfold_data,params,k,a_calib)
    
    if average == True:
        unfold = average_unfold_data(unfold, params)
    
    if plotVerbose == True :
        import pylab
        pylab.hold(False)
        if approach != None:
            pylab.plot(approach['Z_piezo_out'],approach['Deflection_in'], 'b-')
        pylab.hold(True)
        pylab.plot(unfold['Z_piezo_out'], unfold['Deflection_in'], 'g-')
        pylab.show()
    return (approach, unfold, params)

def scale_file(raw_path, target_dir=None, average=True) :
    raw_dir,raw_name = os.path.split(raw_path)
    if VERBOSE >= 1:
        print "working on '%s'" % raw_name
    if target_dir == None :
        target_base = data_logger.normalize_logdir(LOG_DIR)
        raw_basedir,final_dir = os.path.split(raw_dir)
        target_dir = os.path.join(target_base, final_dir)    
    # read in, scale, and average the data
    approach, unfold_avg, params = read_scale_avg(raw_path, average=average)
    # save the data as ASCII arrays
    targ_unfold = os.path.join(target_dir, raw_name)
    dataFile = open(targ_unfold, "w")
    for i in range(0, len(unfold_avg["Z_piezo_out"])) :
        dataFile.write("%g\t%g\n" % (unfold_avg["Z_piezo_out"][i],
                                     unfold_avg["Deflection_in"][i]))
    dataFile.close()
    if approach != None:
        dataFile = open(targ_unfold+"_approach", "w")
        for i in range(0, len(approach["Z_piezo_out"])) :
            dataFile.write("%g\t%g\n" % (approach["Z_piezo_out"][i],
                                         approach["Deflection_in"][i]))
        dataFile.close()
    # and copy the parameters file
    shutil.copy2(raw_path+"_param", target_dir)
    new_param_path = os.path.join(target_dir, raw_name+"_param")
    pfile = file(new_param_path, "a")
    pfile.write("Scaled data fields:\tZ_piezo_out (nm)\tDeflection_in (pN)\n")
    pfile.write("\n")
    pfile.write("Scaling parameters\n")
    pfile.write("Cantilever calibration file: %s\n"
                % params['Cantilever calibration file'])
    pfile.write("Average: %s\n" % average)
    pfile.write("Invert Z: %s\n" % INVERT_Z)
    pfile.write("Remove cantilever displacement: %s\n" % REMOVE_CANT)
    pfile.write("Use calibration photodiode sensitivity: %s\n"
                % CALIB_SENSITIVITY)
    pfile.write("Zero deflection from tailing N: %s\n"
                % ZERO_DEF_FROM_TAILING_N)
    pfile.write("Allow missing approach: %s\n" % ALLOW_MISSING_APPROACH)
    pfile.close()

def scale_directory(raw_data_dir, target_base) :
    """
    Read all the unfolding data from the dir,
    and save scaled data to target_base/final_dir.
     e.g.
      raw_data_dir : ~/rsrch/data/unfold/20080110
      target_base  : ~/rsrch/data/scaled_unfold
     will take all '*_unfold'-based file groups from raw_data_dir,
     and place them in
      ~rsrch/data/scaled_unfold/20080110
    The final directory (20080110 here) can be any valid filename.
    The output files are tab ('\t') delimited ASCII text arrays:
     *_unfold             (scaled and averaged)
     *_unfold_approach    (scaled)
    And the colon (':') delimited ASCII parameters file:
     *_unfold_param       (copied)
    """
    target_base = data_logger.normalize_logdir(target_base)
    # get the name of the final directory and create it if neccessary
    head, final_dir = os.path.split(raw_data_dir)
    # if raw_data_dir ends in a slash, basename() returns empty string
    while final_dir == "" : # keep pulling off trailing slashes until we get a dir.
        head, final_dir = os.path.split(head)
        
    target_dir = os.path.join(target_base, final_dir)
    if os.path.exists(target_dir) :
        pass # Note: we may clobber any files already inside this dir
    else :
        if not os.path.exists(target_base) :
            os.mkdir(target_base, 0755)
        os.mkdir(target_dir, 0755)

    # get a list of unfold datafiles to scale.
    raw_files = list_re_search(os.listdir(raw_data_dir), ".*_unfold$")

    # and do it ;)
    for raw_name in raw_files :
        raw_path = os.path.join(raw_data_dir, raw_name)
        try :
            scale_file(raw_path, average=AVERAGE)
        except z_piezo_utils.poorFit, string :
            print "Ignoring error: ", string
            continue # move on to next path

def main():
    import sys
    from optparse import OptionParser

    global LOG_SCALED_AVGD_DATA, \
        LOG_DIR, \
        CANTILEVER_CALIBRATION_DIR, \
        FORCED_CANTILEVER_CALIBRATION_FILE, \
        SET_ZERO_POS, \
        AVERAGE, \
        INVERT_Z, \
        INVERT_F, \
        REMOVE_CANT, \
        CALIB_SENSITIVITY, \
        ZERO_DEF_FROM_TAILING_N, \
        ALLOW_MISSING_APPROACH, \
        VERBOSE
    
    usage_string = """usage: %prog [options] data_directory [more directories]

2008, W. Trevor King.
"""
    
    parser = OptionParser(usage=usage_string)
    parser.add_option("-o", "--output-dir", dest="output_dir",
                      help="write output to directory DIR (%default)",
                      metavar="DIR", default=LOG_DIR)
    parser.add_option("-c", "--calibration-dir", dest="calib_dir",
                      help="search for cantilever calibration files in DIR (%default)",
                      metavar="DIR", default=CANTILEVER_CALIBRATION_DIR)
    parser.add_option("-C", "--calibration-file", dest="calib_file",
                      help="override cantilever calibration file search and use FILE",
                      metavar="FILE", default=None)
    parser.add_option("-a", "--average", dest="average", action="store_true",
                      help="average force for each step (%default)",
                      default=AVERAGE)
    parser.add_option("-z", "--invert-z", dest="invert_z", action="store_true",
                      help="invert the z piezo displacement (%default)",
                      default=INVERT_Z)
    parser.add_option("-F", "--invert-F", dest="invert_F", action="store_true",
                      help="invert the cantilever force (%default)",
                      default=INVERT_F)
    parser.add_option("-k", "--remove-cantilever-bending", dest="remove_cant",
                      action="store_true",
                      help="remove cantilever bending distance (%default)",
                      default=REMOVE_CANT)
    parser.add_option("-s", "--calib-sensitivity", dest="calib_sensitivity",
                      action="store_true",
                      help="use the calibrated cantilever sensitivity (as opposed to a sensitivity extracted from the contact region of the curve being scanned) (%default)",
                      default=CALIB_SENSITIVITY)
    parser.add_option("-Z", "--zero-def-from-tailing",
                      dest="zero_def_from_tailing_n", metavar="N", type="int",
                      help="by default, we find the zero deflection by averaging all off-the-surface points.  You may use this option to average only the last N points instead.",
                      default=ZERO_DEF_FROM_TAILING_N)
    parser.add_option("--print-k", dest="print_k",
                      action="store_true",
                      help="Instead of scaling, simply print the appropriate spring constant and exit.  This option only applies to individual files, and only the first file is processed.")
    parser.add_option("--allow-missing-approach",
                      dest="allow_missing_approach",
                      action="store_true",
                      help="Prof. Yang's software does not save approach data (so far as I know to date).  This option allows you to move on without it.")
    parser.add_option("-v", "--verbose",
                      action="count", dest="verbose", default=VERBOSE,
                      help="increment verbosity")
    
    (options, args) = parser.parse_args()

    LOG_DIR = options.output_dir
    CANTILEVER_CALIBRATION_DIR = options.calib_dir
    FORCED_CANTILEVER_CALIBRATION_FILE = options.calib_file
    AVERAGE = options.average
    INVERT_Z = options.invert_z
    INVERT_F = options.invert_F
    REMOVE_CANT = options.remove_cant
    CALIB_SENSITIVITY = options.calib_sensitivity
    ZERO_DEF_FROM_TAILING_N = options.zero_def_from_tailing_n
    ALLOW_MISSING_APPROACH = options.allow_missing_approach
    VERBOSE = options.verbose

    for path in args :
        if os.path.isdir(path) :
            scale_directory(path, options.output_dir)
        elif os.path.isfile(path) :
            if options.print_k:
                approach_data, unfold_data, params = load_pull(path)
                calibfile = find_previous_calibration_file(params["Time"])
                k = get_k_from_file(calibfile)
                print k, "N/m"
                sys.exit(0)
            else:
                try :
                    scale_file(path, average=options.average)
                except z_piezo_utils.poorFit, string :
                    print "Ignoring error: ", string
                    continue # move on to next path
        else :
            raise Exception, "'%s' does not exist" % path


if __name__ == "__main__" :
    main()
